"""
    crossover!(model, de, group)

Performs crossover step for each particle pt in the chain

# Arguments

- `model`: model containing a likelihood function with data and priors
- `de`: differential evolution object
- `group`: a group of particles
"""
function crossover!(model, de, group)
    for pt in group
        if rand() ≤ de.θsnooker
            # generate the proposal
            proposal,log_adj = snooker_update!(de, pt, group)
            # compute the weight of the proposal: prior loglikelihood + data loglikelihood
            de.evaluate_fitness!(de, model, proposal)
            # accept proposal according to Metropolis-Hastings rule
            de.update_particle!(de, pt, proposal, log_adj)
        else
            # generate the proposal
            proposal = de.generate_proposal(de, pt, group)
            # compute the weight of the proposal: prior loglikelihood + data loglikelihood
            de.evaluate_fitness!(de, model, proposal)
            # accept proposal according to Metropolis-Hastings rule
            de.update_particle!(de, pt, proposal)
        end
    end
    return nothing
end

"""
    resample(de, group, n, replace)

Sample a random particle from previously accepted values for snooker update.

# Arguments

- `de`: differential evolution object
- `group`: a group of particles
- `n`: number of particles to sample
- `_`: sample with replacement if true (not used)
"""
function resample(de, group, n, _)
    P′ = Vector{eltype(group)}(undef,n)
    mx_idx = de.iter - 1
    for i in 1:n 
        idx = rand(1:mx_idx)
        j = rand(1:length(group))
        P′[i] = Particle(;Θ=group[j].samples[idx,:])
    end
    return P′
end

"""
    sample(de::DE, group_diff, n, replace) 

Sample a random particle.

# Arguments

- `de`: differential evolution object
- `group`: a group of particles
- `n`: number of particles to sample
- `replace`: sample with replacement if true
"""
function sample(de::DE, group_diff, n, replace) 
    return sample(group_diff, n; replace)
end

"""
    random_gamma(de, Pt, group)

Generate proposal according to θ' = θt + γ1(θm − θn) + γ2(θb − θt) + b.
γ2=0 after burnin

# Arguments

- `de`: differential evolution object
- `Pt`: current particle
- `group`: a group of particles
"""
function random_gamma(de, Pt, group)
    Np = length(Pt.Θ)
    # select the base particle θb
    Pb = select_base(group)
    # group without Pb
    group_diff = setdiff(group, [Pt])
    # sample particles for θm and θn
    Pm,Pn = de.sample(de, group_diff, 2, false) 
    # sample gamma weights
    γ₁ = rand(Uniform(.5, 1))
    # set γ₂ = 0 after burnin
    γ₂ = de.iter > de.burnin ? 0.0 : rand(Uniform(.5, 1))
    # sample noise for each parameter
    b = Uniform(-de.ϵ, de.ϵ)
    # compute proposal value
    Θp = Pt + γ₁ * (Pm - Pn) + γ₂ * (Pb - Pt) + b
    # reset each parameter to previous value with probability (1-κ)
    recombination!(de, Pt, Θp)
    return Θp
end

"""
    fixed_gamma(de, Pt, group)

Generate proposal according to θ' = θt + γ(θm − θn) + b
where γ = 2.38.

# Arguments

- `de`: differential evolution object
- `Pt`: current particle
- `group`: a group of particles
"""
function fixed_gamma(de, Pt, group)
    Np = length(Pt.Θ)
    group_diff = setdiff(group, [Pt])
    # sample particles for θm and θn
    Pm,Pn = de.sample(de, group_diff, 2, false) 
    # sample gamma weights
    γ = 2.38
    # sample noise for each parameter
    b = Uniform(-de.ϵ, de.ϵ)
    # compute proposal value
    Θp = Pt + γ * (Pm - Pn) + b
    # reset each parameter to previous value with probability (1-κ)
    recombination!(de, Pt, Θp)
    return Θp
end

"""
    variable_gamma(de, Pt, group)

Generate proposal according to θ' = θt + γ(θm − θn) + b
where γ = 2.38/√(2d) where d is the number of parameters

# Arguments 

- `de`: differential evolution object
- `Pt`: current particle
- `group`: a group of particles
"""
function variable_gamma(de, Pt, group)
    Np = length(Pt.Θ)
    group_diff = setdiff(group, [Pt])
    # sample particles for θm and θn
    Pm,Pn = de.sample(de, group_diff, 2, false) 
    γ = 2.38 / sqrt(2 * Np)
    # sample noise for each parameter
    b = Uniform(-de.ϵ, de.ϵ)
    # compute proposal value
    Θp = Pt + γ * (Pm - Pn) + b
    # reset each parameter to previous value with probability (1-κ)
    recombination!(de, Pt, Θp)
    return Θp
end

"""
    snooker_update!(de, Pt, group)

Performs snooker update during crossover

# Arguments 

- `de`: differential evolution object
- `Pt`: current particle
- `group`: a group of particles
"""
function snooker_update!(de, Pt, group)
    Np = length(Pt.Θ)
    Pz,Pm,Pn = de.sample(de, group, 3, false) 
    Pd = Pt - Pz
    Pr1 = project(Pm, Pd)
    Pr2 = project(Pn, Pd)
    γ = rand(Uniform(1.2, 2.2))
    # sample noise for each parameter
    b = Uniform(-de.ϵ, de.ϵ)
    # compute proposal value
    Θp = Pt + γ * (Pr1 - Pr2) + b
    # reset each parameter to previous value with probability (1-κ)
    recombination!(de, Pt, Θp)
    adj1 = norm(Θp - Pz)^(Np - 1)
    adj2 = norm(Pt - Pz)^(Np - 1)
    return Θp,log(adj1 / adj2)
end

"""
    select_base(group)
Selects base particle θb with probability proportional to weight.

- `group`: a group of particles
"""
function select_base(group)
    w = map(x -> x.weight, group)
    θ = exp.(w) / sum(exp.(w))
    p = sample(group, Weights(θ))
    return p
end

"""
    recombination!(de, pt::Particle, pp::Particle)

Resets parameters of proposal to previous value with probability
(1-κ).

- `de`: differential evolution object
- `pt`: current partical
- `pp`: proposal particle
"""
function recombination!(de, pt::Particle, pp::Particle)
    de.κ == 1.0 ? (return nothing) : nothing
    N = length(pt.Θ)
    for i in 1:N
        if isa(pp.Θ[i], Array)
            recombination!(de, pt.Θ[i], pp.Θ[i])
        else
            pp.Θ[i] = rand() <= (1 - de.κ) ?  pt.Θ[i] : pp.Θ[i]
        end
    end
    return nothing
end

# Handles elements within an array
function recombination!(de, Θt, Θp)
    N = length(Θt)
    for i in 1:N
        Θp[i] = rand() <= (1 - de.κ) ? Θt[i] : Θp[i]
    end
    return nothing
end